(** Directories *)

class ['a,'b] directory = object (self)

  val items : (string,'a) Hashtbl.t = Hashtbl.create 11
  val mutable subdirs : (string,'b) Hashtbl.t = Hashtbl.create 11

  (* Write *)
  method add_item name x = Hashtbl.add items name x
  method add_directory name x = Hashtbl.add subdirs name x

  (* Read *)
  method iter f =
    Hashtbl.iter (fun _ v -> f v) items ;
    Hashtbl.iter (fun _ d -> d#iter f) subdirs

end

class type [+'a] ro_directory = object
  method iter : ('a -> unit) -> unit
end

type 'a dir = ('a, 'a ro_directory) directory

(** Pretty printers *)

class type pretty_printer = object
  method open_section : string -> unit
  method close_section : unit
  method paragraph : string -> unit
end

class terminal_pp = object
  val mutable indent = 0
  method open_section name =
    for i = 1 to indent do Printf.printf " " done ;
    Printf.printf "%s\n" name ;
    indent <- indent+1
  method close_section =
    indent <- indent-1
  method paragraph text =
    for i = 1 to indent do Printf.printf " " done ;
    Printf.printf "%s\n" text
end

(* Extension to printable directories *)

class ['a,'b] printable_directory = object
  inherit ['a,'b] directory
  method print (pp:pretty_printer) =
    Hashtbl.iter (fun name i -> pp#paragraph (name ^ ": " ^ i#to_string)) items ;
    Hashtbl.iter
      (fun name d -> pp#open_section name ; d#print pp ; pp#close_section)
      subdirs
end

class type [+'a] printable_ro_directory = object
  method print : pretty_printer -> unit
  method iter : ('a -> unit) -> unit
end

type 'a pdir = ('a, 'a printable_ro_directory) printable_directory

let create_printable_directory () = new printable_directory
